/* Design of FIR filter for DSP-10 receive audio using window method.
 * Routines for downloading to DSP in DSP-10 project
 *
 * RSL   30 Nov 1998, Initial write.
 * RSL   1 July 2000, Added equalizer.
 * RSL   17 Dec 2000, Added graphics and modal dialog box
 * RSL   2 May 2001   Brought comments up to readable level
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <graphics.h>
#include <string.h>
#include <dos.h>
#include <bios.h>
#include <conio.h>
#include <time.h>
#include <alloc.h>
#include "async.h"
#include "lbox.h"
#include "moonsun.h"
#include "util.h"
#include "uhfa.h"

/* log2(x)=KLN2*ln(x) where KLN2=1/ln(2) and ln(x) in C syntax is log(x) */
#define  KLN2       1.44269504089

/* The next two items are for plotting the response of the equalizer. */
#define  NFREQ      128             /* 512 */
#define  PIONNFREQ  0.0245436926    /* 0.00613592315 */
/* Note - Because of complications in convolving in an equalizer (future)
 * the filters in the DSP-10 should generally be odd numbers of taps.
 * Because of that, this program only works with odd numbers, saving
 * the chore of checking whether even or odd.
 */
#define  RAUD_TAPS  199            /* Taps in receiver audio FIR */
#define  HTAPS      99             /* (RAUD_TAPS-1)/2 */
/* On the next, reduce very small epsilon to prevent sqrt(neg number) */
#define  SCALE_XI   2.5507601e-5   /* 199 taps, 1/(RAUD_TAPS-1)^2-epsilon */
#define  FSAMPLE    9600.0

int coef_raud_filter
    (double*, double*, int*, double*, double, double, double, int*, int);

double i0(double);
void do_flow(char*);
void do_fhigh(char*);
void do_kaiser_db(char*);
void plot_filter(double*);
void download_fil8(int*);
void show_wait(int, int);

int	Coeffs_16bit[HTAPS+1];
int	filscale;			/* Num bits to scale FIR 0, 1,2,...*/
/*  int	kdb=10;  */

/* ************************************************************************ */
#if 0
main(void)
	{/* For testing these routines */
	int i, j, jj;
	int err;
	int grdrv, grmode;
	char cc[50];

	grdrv=DETECT;	
	detectgraph(&grdrv, &grmode);

	printf("Graphics driver= %d, Mode %d\n", grdrv, grmode);
	if (grdrv != 9)
	  {
	  printf("If driver is not 9 (VGA) & Mode is not 2 (640x480) you may");
	  printf(" encounter\ngraphics driver incompatibilities. VGA will be");
	  printf("initialized after the next return\n");
	  }

	grdrv=9;			/* Only allow VGA 640x480 */
	grmode=2;
	initgraph(&grdrv, &grmode, "");
	setbkcolor(BLUE);
	draw_grid();

	if (err=coef_raud_filter
	    (&feqr[0], &aeqrdb[0], &Coeffs_16bit[0], &rsp_db[0],
		350.0, 1250.0, 70.0, &filscale, 3))
	    printf("\nError %d\n", err);

	/* Plot the response */
	setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
	setcolor(LIGHTRED);
	moveto(0,128);
	moveto(0, -floor(0.5+1.6*rsp_db[0]));
	for (i=1; i<NFREQ; i++)
	    lineto(i, -floor(0.5+1.6*rsp_db[i]) );

	/* Small box response curve, 64 high (80 dB), 128 (4800 Hz) wide. */
	setviewport(500, 200, 639, 275, DO_CLIP);
	clearviewport();
	setcolor(WHITE);
	for (j=0; j<=8; j++)
	    line(0,  8*j, 128,  8*j);
	for (j=0; j<=4; j++)
	    {
	    jj=(int)(0.5+26.6666667*(double)j);
	    line(jj, 0, jj, 64);
	    }
	line(128, 0, 128, 64);
	moveto(0, 66);
	settextjustify(LEFT_TEXT, TOP_TEXT);
	outtext("0     2     4kHz");

	setviewport(582, 202, 626, 212, DO_CLIP);
	clearviewport();
	moveto(43, 3);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	outtext("10dB/");

	setviewport(500, 200, 639, 264, DO_CLIP);
	setcolor(LIGHTRED);
	moveto(0,  -floor(0.5+0.8*rsp_db[0]));
	for (i=1; i<NFREQ/4; i++)
	   lineto(i, -floor(0.5+0.8*rsp_db[4*i]) );
	getch();
	setbkcolor(BLACK);
	setcolor(WHITE);
	closegraph();
	return(0);
	}
#endif


void filter_design_modalbox(void)
	{

	static struct lbox_info   lb1;

	static struct edit_info   el1, el2, el3, el4;
	static struct check_info  ck1;
	char eline1[12], eline2[12], eline3[12], eline4[12];
	unsigned int edresult, rbui, ckui;
	int err;
	double f;
	static double *rsp_db;

	/* Get memory for double response */
	if ( (rsp_db = (double *)calloc(512, sizeof(double)) ) == NULL)
		{
		do_t_error(48, "E48 Dyn Mem-Fil", 20);
		return;
		}

	/* A modal box, 25 characters wide, 25 lines high */
	lb_create(&lb1, "RECEIVE FILTER #8", 160, 175, 370, 440, 0, MODAL);
	lb_open(&lb1, 1);
	/* Adding 64+lines, after lb_create() leaves graphics area uncleared */
	lb1.lines |= 0X0040;

	/* Fixed text in modal box:     */
	lb_line(&lb1,  9,  TXT, "Kaiser Filter + Equalizer", 0);
	lb_line(&lb1, 21, BOLD, "\030 and \031 Move to Next Item", 0);
	lb_line(&lb1, 22, BOLD, "SPACE=Change Button/Box", 0);
	lb_line(&lb1, 23, BOLD, "TEXT: Backspace & Type", 0);
	lb_line(&lb1, 24, BOLD, "ENTER=Close Dialog", 0);

	/* Use filter check box:     */
	if (fil8_use)			/* Show box checked */
	    lb_create_check(&ck1, 0, TRUE, 20);
	else				/* Show box unchecked */
	    lb_create_check(&ck1, 0, FALSE, 20);
	lb_line(&lb1, 19, CHECK, "Use Filter on Exit", &ck1);

	/* Editable text areas:         */
	/* Note edit boxes take a line and the next one, ie el1 uses 4 and 5 */
	sprintf(eline1, "%d", (int)(0.5+flow));
	lb_create_el(&el1, 5, 143, &eline1[0], ED_INACTIVE, 0, strlen(eline1));
	lb_line(&lb1, 11, EDIT, " Lower Pass-band", &el1);

	sprintf(eline2, "%d", (int)(0.5+fhigh));
	lb_create_el(&el2, 5, 143, &eline2[0], ED_INACTIVE, 0, strlen(eline2));
	lb_line(&lb1, 13, EDIT, " Upper Pass-band", &el2);

	sprintf(eline3, "%d", (int)(0.5+kaiser_db));
	lb_create_el(&el3, 3, 159, &eline3[0], ED_INACTIVE, 0, strlen(eline3));
	lb_line(&lb1, 15, EDIT, " 1st Sidelobe, dB", &el3);

	sprintf(eline4, "%s", fil8_name);
	lb_create_el(&el4, 7, 127, &eline4[0], ED_INACTIVE, 0, strlen(eline4));
	lb_line(&lb1, 17, EDIT, " Filter Name", &el4);

	lb_update(&lb1);

 e11:	show_wait(220, 220);
	if ((err=coef_raud_filter
	    (&feqr[0], &aeqrdb[0], &Coeffs_16bit[0], &rsp_db[0],
		flow, fhigh, kaiser_db, &filscale, 3))==TRUE)
	    do_t_error(49, "E49 Math-Filt", 20);
	plot_filter(&rsp_db[0]);
	edresult=lb_edit_line(&lb1, &el1);
	do_flow(eline1);
	if (edresult==CUR_UP)   goto c19;
	else if (edresult==ENTER) goto lb1end;

 e13:	show_wait(220, 220);
	if ((err=coef_raud_filter
	    (&feqr[0], &aeqrdb[0], &Coeffs_16bit[0], &rsp_db[0],
		flow, fhigh, kaiser_db, &filscale, 3))==TRUE)
	    do_t_error(49, "E49 Math-Filt", 20);
	plot_filter(&rsp_db[0]);
	edresult=lb_edit_line(&lb1, &el2);
	do_fhigh(eline2);
	if (edresult==CUR_UP)   goto e11;
	else if (edresult==ENTER) goto lb1end;

 e15:	show_wait(220, 220);
	if ((err=coef_raud_filter
	    (&feqr[0], &aeqrdb[0], &Coeffs_16bit[0], &rsp_db[0],
		flow, fhigh, kaiser_db, &filscale, 3))==TRUE)
	    do_t_error(49, "E49 Math-Filt", 20);
	plot_filter(&rsp_db[0]);
	edresult=lb_edit_line(&lb1, &el3);
	do_kaiser_db(eline3);
	if (edresult==CUR_UP)   goto e13;
	else if (edresult==ENTER) goto lb1end;

 e17:	show_wait(220, 220);
	if ((err=coef_raud_filter
	    (&feqr[0], &aeqrdb[0], &Coeffs_16bit[0], &rsp_db[0],
		flow, fhigh, kaiser_db, &filscale, 3))==TRUE)
	    do_t_error(49, "E49 Math-Filt", 20);
	plot_filter(&rsp_db[0]);
	edresult=lb_edit_line(&lb1, &el4);
	if (edresult==CUR_UP)   goto e15;
	else if (edresult==ENTER) goto lb1end;

 c19:	ckui=lb_check_mark(&lb1, &ck1);
	if     (ckui==CUR_UP)  goto e17;
	else if (ckui==CUR_DOWN) goto e11;
	else if(ckui==ENTER)   goto lb1end;

lb1end: /* As we exit, the variables must be changed to agree with the
	 * indicated states of the buttons and edit fields:
	 */
	show_wait(220, 220);
	if (ck1.state & CHECK_MARK)
	    fil8_use=TRUE;
	else
	    fil8_use=FALSE;
	do_flow(eline1);
	do_fhigh(eline2);
	do_kaiser_db(eline3);
	strcpy(fil8_name, eline4);

	if ((err=coef_raud_filter
	    (&feqr[0], &aeqrdb[0], &Coeffs_16bit[0], &rsp_db[0],
		flow, fhigh, kaiser_db, &filscale, 3))==TRUE)
	    do_t_error(49, "E49 Math-Filt", 20);
	plot_filter(&rsp_db[0]);

	if(fil8_use)
	    {
	    download_fil8(&Coeffs_16bit[0]);
	    /* Now mark the info in the filt[7] srtructure */
	    strcpy(filt[7].name, fil8_name);
	    filt[7].fil_type=1;		/* Kaiser */
	    filt[7].param1=kaiser_db;
	    filt[7].fclow=flow;
	    filt[7].fchigh=fhigh;
	    filt[7].fil_gain=filscale;
	    af_coef=7;			/* And select filter 8 */
	    set_affilt_select();
	    }
	else
	    lb_msg(170, 260, "INFORMATION",
		" 'Use filter on Exit' was not",
		" Checked. Filter not loaded. ");
	delay(500);
	lb_close(&lb1, 1);
	free(rsp_db);
	}

void do_flow(char *el)
	{
	double f;
	f=atof(el);
	if (f<0.0)
	    {
	    f=0.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies must be positive",
		"   Lower Passband set to 0");
	    }
	else if (f>4400.0)
	    {
	    f=4300.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies  must be < 4400",
		" Lower Pass-Band set to 4300");
	    }
	flow=f;				/* flow is global */
	}

void do_fhigh(char *el)
	{
	double f;
	f=atof(el);
	if (f<0.0)
	    {
	    f=0.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies must be positive",
		"   Upper Pass-band set to 0");
	    }
	else if (f>4400.0)
	    {
	    f=4400.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies  must be < 4400",
		" Upper Pass-band set to 4399");
	    }
	if (f < flow)
	    {
	    f=flow+10.0;
	    lb_msg(170, 260, "WARNING",
		"Lower must be less than Upper",
		"  Upper Pass-band adjusted");
	    }
	fhigh=f;
	}

void do_kaiser_db(char *el)
	{
	double db;

	db=atof(el);
	if (db<0.0)
	    {
	    db=50.0;
	    lb_msg(170, 260, "WARNING",
		" Sidelobe level must be positive",
		"  Sidelobe level set to 50");
	    }
	else if (db>200.0)
	    {
	    db=100.0;
	    lb_msg(170, 260, "WARNING",
		" Sidelobe level must be < 100",
		"  Sidelobe level set to 99");
	    }
	kaiser_db=db;
	}


/* coef_raud_filter() calculates the FIR filter coefficients for
 * receive filter and equalizer.
 * Because of symmetry, only half of the coefficients have been
 * calculated. When loading to the DSP, this needs to be allowed for.
 *
 *   feq[]    Array of NFRB (15) breakpoint frequencies, fractions of 9.6 kHz.
 *   aeq[]    Array of NFRB-1 levels for the feq[] defined frequency bands.
 *   cf16[]   Array of (RAUD_TAPS+1)/2 FIR coefficients (returned).
 *   rdb[]    Array of NFREQ db levels that is response of filter (returned).
 *
 * Returns 0 if successful, or an error code if not.
 * Errors:  1=Unable to obtain memory array
 *          2=Illegal bandpass limits
 */
int coef_raud_filter
    (double* feq, double* adb, int* cf16, double* rdb,
     double fl, double fh, double kdb,
     int* pscale, int scalemax)
	{
	int i, j, jj, sci;
	double q, kbes, sc;
	double temp;
	double bt, btmax;
	double xi, beta, WindowWt;
	double *cf;
	double cfmax;
	double fl0, fh0;			/* Local copies, / FSAMPLE */
	static double ae[NFRB-1];		/* Equalizer level V ratios */
	static double f[NFRB], a[NFRB-1];	/* Combined filt + equalizer */


	/* Get memory for double coefficients */
	if ( (cf = (double *)calloc(HTAPS+1, sizeof(double)) ) == NULL)
		 return 1;

	/* Convert from dB to voltage ratios, initialize local arrays */
	for (i=0; i<NFRB-1; i++)
	    {
	    ae[i]=pow(10.0, (0.05*adb[i])); /* Equalizer response */
	    a[i]=ae[i];			/* Local version we can modify */
	    f[i]=feq[i];		/* Again, local versions */
	    }
	f[NFRB-1]=feq[NFRB-1];		/* Missed one! */

	/* We need a parameter beta for the Kaiser window. It is a function
	 * of the kdb sidelobe level. See IEEE book.
	 */
	if (kdb>50)
	    beta = 0.1102*(kdb-8.7);
	else if  (kdb>20.96 && kdb<=50.0)
	    beta = 0.58417*pow((kdb-20.96), 0.4) + 0.07886*(kdb-20.96);
	else
	    beta=0.0;
	kbes=1.0/i0(beta);		/* A derived parameter used in loop */

	/* Make sure values are credible */
	if (fl>=fh || fl<0.0 || fh>FSAMPLE*feq[NFRB-1])
	   {
	   free(cf);
	   return 2;
	   }

	fl0=fl/FSAMPLE;			/* Normalized copies */
	fh0=fh/FSAMPLE;

	/* Overlay the (fl, fh) bandpass filter on the equalizer response: */
	for(i=1; i<NFRB-2; i++)		/* Start from the bottom for fl */
	    if(fl0 > f[i])   a[i-1]=0.0;
	for(i=NFRB-2; i>=1; i--)		/* Start from the top for fh */
	    if(fh0 < f[i])   a[i]=0.0;
	for(i=NFRB-2; i>=0; i--)
	    if(fl0 > f[i])   f[i]=fl0;
	for(i=0; i<NFRB; i++)
	    if(fh0 < f[i])   f[i]=fh0;

	/* Find FIR coefficients, the Fourier transform of the frequency
	 * response. This is done by dividing the response into a sequence
	 * of NFRB-1 rectangular frequency blocks, each of a different level.
	 * Linearity of the Fourier transform allows us to sum the transforms
	 * of the individual blocks.
	 */
	/* First the center coef  (for 199 taps, this is [99] */
	cf[HTAPS] = 2.0*a[NFRB-2]*f[NFRB-1]-2.0*(a[0]*f[0]);
	for(i=0; i<NFRB-2; i++)
	   cf[HTAPS] += 2.0*a[i]*f[i+1]-2.0*(a[i+1]*f[i+1]);

	for (j=0; j<HTAPS; ++j)		/* 0 to 98 */
	    {
	    q = PI*(double)(j-HTAPS);

	    /* First, deal with the end points */
	    cf[j] = -a[0]*sin(f[0]*2.0*q)/q
	            +a[NFRB-2]*sin(f[NFRB-1]*2.0*q)/q;
	    /*   and then the intermediate points */
	    for(i=0; i<NFRB-2; i++)
		cf[j] += (a[i]-a[i+1])*sin(f[i+1]*2.0*q)/q;
	    }

	for (j=0; j<=HTAPS; j++)	/* For 199 Taps, this is 0 to 99 */
	    {xi=(int)(0.5+(double)j);
	     xi=4.0*xi*xi;
	     WindowWt=kbes*(i0(beta*sqrt(1.0-SCALE_XI*xi))); /*Kaiser win*/
	     /* Up to here the center of the filter has been cf[0], now flip */
	     jj=HTAPS-j;
	     cf[jj] *= WindowWt;	/* and apply the Kaiser window */
	    }

	/* Compute response to normalize the response to 1.0 at maximum.
	 * For linear phase, the filter is symmetric about the center. Thus
	 * the response is twice the response of the first half. For odd
	 * number of taps, the center tap response is a special case and
	 * is also multiplied by 0.5 since it is part of both halves.
	 */
	btmax=0.0;
	for (i=0; i<NFREQ; i++)
	    {
	    bt = 0.5*cf[HTAPS];		/* for odd number of taps */
					/* bt=0.0 for even number */
	    for (j=0; j<HTAPS; j++)
		bt += cf[j]*cos(PIONNFREQ*(double)((HTAPS-j)*i));

	    bt=fabs(bt);
	    if(bt>btmax)  btmax=bt;
	    }
	btmax *= 2.0;			/* Because based on half of coef */

	/* Normalize the response to maximum gain of 1.0 */
	for (j=0; j<=HTAPS; j++)
	    {
	    cf[j] /= btmax;
	    }

	/* Find biggest to decide the scaling factor for the FIR filter.
	 * Narrow band filters produce small coefficient values that have
	 * excessive quantization noise. The FIR filter can be allowed
	 * to temporarily overrange to minimize this effect. Here we
	 * determine the scaling factors.
	 */
	cfmax=0.0;
	for (j=0; j<=HTAPS; j++)		/* 0 to 99 */
	    if (cfmax < fabs(cf[j])) cfmax=fabs(cf[j]);

	sci=-(int)(KLN2*log(cfmax));	/* If cfmax=.1, sci=3, etc */

	if (sci > scalemax) sci=scalemax;
	*pscale = sci;			/* Output the scale value */

	sc=pow(2, (double)sci);	/* To scale cf[] up, closer to 1.0 */

	/* Adjust the coefficients up towards 1 and convert to 16 bit:    */
	for (j=0; j<=HTAPS; j++)
	    {
	    cf[j] *= sc;
	    cf16[j] = (int)(32768.0*cf[j]);
	    }

	/* Calculate response for plotting in top box.
	 * Convert back to double to include effects of quantizing.
	 */
	for (j=0; j<=HTAPS; j++)	/* 0 to 99 */
	    cf[j] = (double)cf16[j]/(32768.0);

	for (i=0; i<NFREQ; i++)
	    {
	    bt = 0.5*cf[HTAPS];
	    for (j=0; j<HTAPS; j++)
		bt += cf[j]*cos(PIONNFREQ*(double)((HTAPS-j)*i));

	    /* Add 6 dB since we are based on half of the symmetric cf[].
	     * Scale back by 6.02*sci
	     * as this is the overscale factor.
	     */
	    if (fabs(bt) > 5.01187E-6)
		rdb[i]=6.02-6.02*((double)sci)+20.0*log10(fabs(bt));
	    else
		rdb[i]=-100.0;
	    }
	free(cf);
	return 0;
	}
/* plot_filter(pointer to array of doubles )
 * Small box response curve, 64 high (80 dB), 128 (4800 Hz) wide.
 */
void plot_filter(double *ydb)
	{
	int j, jj;

	#define GRLEFT 199		/* Graph location, absolute coord */
	#define GRTOP  188

	setviewport(GRLEFT, GRTOP, GRLEFT+139, GRTOP+75, DO_CLIP);
	clearviewport();
	setcolor(c_text_bold);
	for (j=0; j<=8; j++)
	    line(0,  8*j, 128,  8*j);
	for (j=0; j<=4; j++)
	    {
	    jj=(int)(0.5+26.6666667*(double)j);
	    line(jj, 0, jj, 64);
	    }
	line(128, 0, 128, 64);
	moveto(0, 66);
	settextjustify(LEFT_TEXT, TOP_TEXT);
	outtext("0     2     4kHz");

	setviewport(GRLEFT+82, GRTOP+2, GRLEFT+126, GRTOP+12, DO_CLIP);
	clearviewport();
	moveto(43, 3);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	outtext("10dB/");

	setviewport(GRLEFT, GRTOP, GRLEFT+139, GRTOP+64, DO_CLIP);
	setcolor(c_plot_marker);
	moveto(0,  -floor(0.5+0.8*ydb[0]));
	for (j=1; j<NFREQ; j++)
	   lineto(j, -floor(0.5+0.8*ydb[j]) );


	setcolor(c_text_bold);
	}


void download_fil8(int *cf16)
	{
	int i, adr1, adr2;

	for(i=0; i<HTAPS; i++)
	    {
	    adr1=dsp_ptr[14]+i;
	    adr2=dsp_ptr[14]+RAUD_TAPS-1-i;
	    ASYNCOUT(ESC_CHAR);  ASYNCOUT(2); /* To PM() */
	    ASYNCOUT(adr1&0XFF);	/* Address low */
	    ASYNCOUT((adr1>>8)&0XFF);	/* Address high */
	    ASYNCOUT(cf16[i]&0XFF);	/* Data middle byte */
	    ASYNCOUT((cf16[i]>>8)&0XFF); /* Data high byte */

	    ASYNCOUT(ESC_CHAR);  ASYNCOUT(2); /* To PM() */
	    ASYNCOUT(adr2&0XFF);	/* Address low */
	    ASYNCOUT((adr2>>8)&0XFF);	/* Address high */
	    ASYNCOUT(cf16[i]&0XFF);	/* Data middle byte */
	    ASYNCOUT((cf16[i]>>8)&0XFF); /* Data high byte */
	    }
	/* Due to odd number of coefficients in FIR, there is center to load */
	adr1=dsp_ptr[14]+HTAPS;
	ASYNCOUT(ESC_CHAR);  ASYNCOUT(2); /* To PM() */
	ASYNCOUT(adr1&0XFF);		/* Address low */
	ASYNCOUT((adr1>>8)&0XFF);	/* Address high */
	ASYNCOUT(cf16[HTAPS]&0XFF);	/* Data middle byte */
	ASYNCOUT((cf16[HTAPS]>>8)&0XFF); /* Data high byte */
	beep(100);
	}

void show_wait(int x, int y)
	{
	setviewport(x, y, x+56, y+9, DO_CLIP);
	clearviewport();
	setcolor(c_text_bold);
	moveto(0, 0);
	settextjustify(LEFT_TEXT, TOP_TEXT);
	outtext("Wait...");
	}

/* i0(x)  Returns the modified Bessel function Io(x).
 * Algorithm is based on Abromowitz and Stegun, Handbook of Mathematical
 * Functions, and Press, et. al., Numerical Recepies in C.
 */
double i0(double x)
	{
	double mdx;
	double bess;
	double poly;
	if( (mdx=fabs(x)) < 3.75 )
	    {
	    poly=x/3.75;
	    poly*=poly;
	    bess=1.0+poly*(3.5156229
		+poly*(3.0899424
		+poly*(1.2067492
		+poly*(0.2659732
		+poly*(0.360768e-1
		+poly*0.45813e-2)))));
	    }
	else
	    {
	    poly = 3.75/mdx;
	    bess=(exp(mdx)/sqrt(mdx))*(0.39894228
		+poly*(0.1328592e-1
		+poly*(0.225319e-2
		+poly*(-0.157565e-2
		+poly*(0.916281e-2
		+poly*(-0.2057706e-1
		+poly*(0.2635537e-1
		+poly*(-0.1647633e-1
		+poly*0.392377e-2))))))));
	    }
	return bess;
	}
